#ifndef CPSE_MATRIX_TPP
#define CPSE_MATRIX_TPP

#include <ostream>

#include <cpse/static.h>
#include <cpse/math/matrix.h>
#include <cpse/math/mathops.h>
#include <cpse/math/utility.h>

namespace cpse {
namespace math {

// expression templates description:
// struct _name_ {
//	static const int rows;
//		<number of rows of the matrix that may be the result of
//		 the expression. a large negative number otherwise>
//	static const int columns;
//		<number of columns of the matrix that may be the result of
//		 the expression. a large negative number otherwise>
//	static const bool is_matrix;
//		<can the expression be evaluated to a matrix?>
//	static const bool is_scalar;
//		<can the expression be evaluated to a scalar?>
//	static const bool error;
//		<propagates detection of errors to the topmost expression>
//	static const bool element_op;
//		<can the expression be evaluated correctly and efficiently
//		 looking at a single element at a time?>
//	typedef base_type;
//		<this type>
//	const _type1_ a; <operand 1 if any>
//	const _type2_ b; <operand 2 if any>
//	_name_(); <empty default constructor>
//	_name_([_type1_, [_type2_]]); <initialize operands>
//	template<int Index> float element() const;
//		<if element_op. evaluate element at Index>
// };

// float constant expression
struct float_keeper {
	static const int rows = -99;
	static const int columns = -99;
	static const bool is_matrix = false;
	static const bool is_scalar = true;
	static const bool error = false;
	static const bool element_op = true;
	typedef float_keeper base_type;
	float f;
	float_keeper() {}
	float_keeper(float v) : f(v) {}
	template<class T> float_keeper(const T &);
	template<int> float element() const { return f; }
};

struct axis_type {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_matrix = true;
	static const bool is_scalar = false;
	static const bool error = false;
	static const bool element_op = true;
};

struct x_axis_type : axis_type {
	typedef x_axis_type base_type;
	template<int I> static float element() { return I == 0; }
};
struct y_axis_type : axis_type {
	typedef y_axis_type base_type;
	template<int I> static float element() { return I == 1; }
};
struct z_axis_type : axis_type {
	typedef z_axis_type base_type;
	template<int I> static float element() { return I == 2; }
};

// fake vector constants
const tmatrix<x_axis_type> x_axis = tmatrix<x_axis_type>();
const tmatrix<y_axis_type> y_axis = tmatrix<y_axis_type>();
const tmatrix<z_axis_type> z_axis = tmatrix<z_axis_type>();

// find out which type should be used to hold references to expressions
// (const basic_matrix &, all others as themselves)
template<class T> struct matrix_ref { typedef const T type; };
template<int A, int B> struct matrix_ref<basic_matrix<A, B> >
	{ typedef const basic_matrix<A, B> &type; };

template<class T, class U, bool A, bool B>
struct select_matrix_helper { typedef T type; };
template<class T, class U, bool B>
struct select_matrix_helper<T, U, false, B> { typedef U type; };

// select the leftmost argument that is a matrix
template<class T, class U>
struct select_matrix {
	typedef typename select_matrix_helper<T, U, T::is_matrix, 
		U::is_matrix>::type type;
};

// for convenience. check if argument is a 3x1 matrix
template<class T> struct is_vector_3
	{ static const bool value = T::rows == 3 && T::columns == 1
		&& T::is_matrix; };

// needed for vector/scalar product resolution. matrix_type is the type of
// the expression if used in a matrix (vector) expression. scalar_type is
// the type if used in a scalar expression. for expressions that could be either
// matrix or scalar the types are different, for all other expressions they are
// the same.
template<class T, bool Is_matrix, bool Is_scalar> struct result_type_helper {};
template<class T> struct result_type_helper<T, true, false> {
	typedef basic_matrix<T::rows, T::columns> matrix_actual;
	typedef basic_matrix<T::rows, T::columns> scalar_actual;
};
template<class T> struct result_type_helper<T, true, true> {
	typedef basic_matrix<T::rows, T::columns> matrix_actual;
	typedef float_keeper scalar_actual;
};
template<class T> struct result_type_helper<T, false, true> {
	typedef float_keeper matrix_actual;
	typedef float_keeper scalar_actual;
};

template<class T> struct result_type
	: result_type_helper<T, T::is_matrix, T::is_scalar> {};

} // namespace math
} // namespace cpse

#include <cpse/math/matrix_unop.tpp>
#include <cpse/math/matrix_binop.tpp>
#include <cpse/math/matrix_operators.tpp>
#include <cpse/math/matrix_assign.tpp>
#include <cpse/math/matrix_inline.tpp>

#endif
